/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include <QtCore/QFile>
#include <QtCore/QFileInfo>

#include "WorkflowRunTask.h"
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowManager.h>
#include <workflow/Schema.h>
#include <workflow_support/SchemaSerializer.h>
#include <workflow_support/WorkflowUtils.h>

#include <core_api/AppContext.h>
#include <core_api/Log.h>
#include <core_api/Settings.h>
#include <core_api/Counter.h>

#include <cmdline/CMDLineRegistry.h>
#include <cmdline/CMDLineUtils.h>
#include <cmdline/CMDLineHelpProvider.h>
#include <cmdline/CMDLineCoreOptions.h>

#define WORKFLOW_RUN_TASK_LOG_CAT "Workflow run task"

namespace GB2 {

static LogCategory log( WORKFLOW_RUN_TASK_LOG_CAT );

/*******************************************
 * WorkflowRunTask
 *******************************************/
WorkflowRunTask::WorkflowRunTask(const Schema& sh, QList<Iteration> lst) :
Task(tr("Execute workflow schema"), TaskFlags(TaskFlag_NoRun) | TaskFlag_ReportingIsSupported) {
    GCOUNTER( cvar, tvar, "WorkflowRunTask" );
    if (lst.isEmpty()) {
        lst.append(Iteration(tr("Main")));
    }
    foreach(const Iteration& it, lst) {
        WorkflowIterationRunTask* t = new WorkflowIterationRunTask(sh, it);
        connect(t, SIGNAL(si_ticked()), SIGNAL(si_ticked()));
        addSubTask(t);
    }
    setMaxParallelSubtasks(MAX_PARALLEL_SUBTASKS_AUTO);
}

QString WorkflowRunTask::generateReport() const {
    QString res;
    res+="<table width='75%'>";
    res+=QString("<tr><th>%1</th><th>%2</th><th>%3</th></tr>").arg(tr("Iterations")).arg(tr("Status")).arg(tr("Details"));
    foreach(Task* sub, getSubtasks()) {
        QString name = Qt::escape(sub->getTaskName());
        QString status = sub->hasErrors() ? tr("Failed") : sub->isCanceled() ? tr("Canceled") : tr("Finished");
        QString error = Qt::escape(sub->getError()).replace("\n", "<br>");
            //AppContext::getTaskScheduler()->getStateName(sub);
        if (sub->hasErrors()) {
            name = "<font color='red'>"+name+"</font>";
            status = "<font color='red'>"+status+"</font>";
        } else if (sub->isCanceled()) {
            status = "<font color='blue'>"+status+"</font>";
        } else {
            status = "<font color='green'>"+status+"</font>";
        }
        res+=QString("<tr><td>%1</td><td>%2</td><td>%3</td></tr>").arg(name).arg(status).arg(error);
    }
    res+="</table>";
    return res;
}

QList<WorkerState> WorkflowRunTask::getState( Actor* actor) {
    QList<WorkerState> ret;
    foreach(Task* t, getSubtasks()) {
        WorkflowIterationRunTask* rt = qobject_cast<WorkflowIterationRunTask*>(t);
        ret << rt->getState(actor);
    }
    return ret;
}

int WorkflowRunTask::getMsgNum( Link* l) {
    int ret = 0;
    foreach(Task* t, getSubtasks()) {
        WorkflowIterationRunTask* rt = qobject_cast<WorkflowIterationRunTask*>(t);
        ret += rt->getMsgNum(l);
    }
    return ret;
}

Task::ReportResult WorkflowRunTask::report() {
    propagateSubtaskError();
    return ReportResult_Finished;
}


/*******************************************
* WorkflowIterationRunTask
*******************************************/
WorkflowIterationRunTask::WorkflowIterationRunTask(const Schema& sh, const Iteration& it) : 
Task(QString("%1").arg(it.name), TaskFlags_NR_FOSCOE), schema(new Schema()), scheduler(NULL) {
    rmap = SchemaSerializer::deepCopy(sh, schema);
    schema->applyConfiguration(it, rmap);
    
    DomainFactory* df = WorkflowEnv::getDomainRegistry()->getById(schema->domain);
    if (!df) {
        stateInfo.setError(  tr("Unknown domain %1").arg(schema->domain) );
        return;
    }
}

WorkflowIterationRunTask::~WorkflowIterationRunTask() {
    lmap.clear();
    DomainFactory* df = WorkflowEnv::getDomainRegistry()->getById(schema->domain);
    if (df) {
        df->destroy(scheduler, schema);
    }
    scheduler = NULL;
    delete schema;
}

void WorkflowIterationRunTask::prepare() {
    if( hasErrors() || isCanceled() ) {
        return;
    }
    
    DomainFactory* df = WorkflowEnv::getDomainRegistry()->getById(schema->domain);
    assert( df != NULL ); // checked in constructor
    foreach(Actor* a, schema->procs) {
        Worker* w = df->createWorker(a);
        if (!w) {
            stateInfo.setError( tr("Failed to create worker %1 %2 in domain %3")\
                .arg(a->getProto()->getId()).arg(a->getId()).arg(schema->domain) );
            return;
        }
    }
    foreach(Link* l, schema->flows) {
        CommunicationChannel* cc = df->createConnection(l);
        if (!cc) {
            stateInfo.setError(  tr("Failed to create connection %1 %2 in domain %3") ); //fixme
            return;
        }
        QStringList lst;
        lst << rmap.key(l->source()->owner()->getId());
        lst << (l->source()->getId());
        lst << rmap.key(l->destination()->owner()->getId());
        lst << (l->destination()->getId());
        QString key = lst.join("|");
        lmap.insert(key, cc);
    }
    scheduler = df->createScheduler(schema, df->getDefaultSettings());
    if (!scheduler) {
        stateInfo.setError(  tr("Failed to create scheduler in domain %1").arg(df->getDisplayName()) );
        return;
    }
    scheduler->init();
    while(scheduler->isReady() && !isCanceled()) {
        Task* t = scheduler->tick();
        if (t) {
            addSubTask(t);
            break;
        }
    }
}

QList<Task*> WorkflowIterationRunTask::onSubTaskFinished(Task* subTask) {
    QList<Task*> tasks;
    if (subTask->hasErrors()) {
        emit si_ticked();
        propagateSubtaskError();
        return tasks;
    }
    while(scheduler->isReady() && !isCanceled()) {
        Task* t = scheduler->tick();
        if (t) {
            tasks << t;
            break;
        }
    }
    emit si_ticked();
    return tasks;
}

Task::ReportResult WorkflowIterationRunTask::report() {
    if (scheduler) {
        scheduler->cleanup();
        if (!scheduler->isDone()) {
            stateInfo.setError( stateInfo.getError() +  tr("\nSchema did not finish.") );
        }
    }
    return ReportResult_Finished;
}

WorkerState WorkflowIterationRunTask::getState(Actor* a)
{
    if (scheduler) {
        return scheduler->getWorkerState(rmap.value(a->getId()));
    }
    return WorkerWaiting;
}

int WorkflowIterationRunTask::getMsgNum( Link* l)
{
    QStringList lst;
    lst << (l->source()->owner()->getId());
    lst << (l->source()->getId());
    lst << (l->destination()->owner()->getId());
    lst << (l->destination()->getId());
    QString key = lst.join("|");
    
    CommunicationChannel* cc = lmap.value(key);
    if (cc) {
        return cc->hasMessage();
    }
    return 0;
}

/*******************************************
* WorkflowRunFromCMDLineTask
*******************************************/
const QString WorkflowRunFromCMDLineTask::WORKFLOW_COMMANDS_PATH_PREFIX = "workflowcommands";

WorkflowRunFromCMDLineTask::WorkflowRunFromCMDLineTask() 
: Task( tr( "Workflow run from cmdline" ), TaskFlag_NoRun ), schema(NULL), optionsStartAt(-1), loadTask(NULL) {
    CMDLineRegistry * cmdLineRegistry = AppContext::getCMDLineRegistry();
    
    // try to process schema without 'task' option (it can only be the first one)
    QStringList pureValues = CMDLineRegistryUtils::getPureValues();
    if( !pureValues.isEmpty() ) {
        QString schemaName = pureValues.first();
        processLoadSchemaTask( schemaName, 1 ); // because after program name
    }
    if( loadTask != NULL ) {
        addSubTask( loadTask );
        return;
    }
    
    // process schema with 'task' option
    int taskOptionIdx = CMDLineRegistryUtils::getParameterIndex( CMDLineCoreOptions::RUN_WORKFLOW );
    if(taskOptionIdx != -1) {
    processLoadSchemaTask( cmdLineRegistry->getParameterValue( CMDLineCoreOptions::RUN_WORKFLOW, taskOptionIdx ), taskOptionIdx );
    }

    if( loadTask == NULL ) {
        setError( tr( "no task to run" ) );
        return;
    }
    addSubTask( loadTask );
}

void WorkflowRunFromCMDLineTask::processLoadSchemaTask( const QString & schemaName, int optionIdx ) {
    loadTask = prepareLoadSchemaTask( schemaName );
    if( loadTask != NULL ) {
        optionsStartAt = optionIdx + 1;
    }
}

LoadWorkflowTask * WorkflowRunFromCMDLineTask::prepareLoadSchemaTask( const QString & schemaName ) {
    QString pathToSchema = getPathToSchemaFile( schemaName );
    if( pathToSchema.isEmpty() ) {
        log.error( tr( "Cannot find schema: %1" ).arg( schemaName ) );
        return NULL;
    }
    
    schema = new Schema();
    schema->deepCopy = true;
    return new LoadWorkflowTask( schema, NULL, pathToSchema );
}

WorkflowRunFromCMDLineTask::~WorkflowRunFromCMDLineTask() {
    delete schema;
}

static Actor * findActorByParamAlias( Schema * schema, const QString & alias, QString & attrName ) {
    assert( schema != NULL );
    QList<Actor*> actors;
    foreach( Actor * actor, schema->procs ) {
        assert( actor != NULL );
        if( actor->getParamAliases().values().contains( alias ) ) {
            actors << actor;
        }
    }
    
    if ( actors.isEmpty() ) {
        return NULL;
    } else if( actors.size() > 1 ) {
        log.error( WorkflowRunFromCMDLineTask::tr( "%1 actors in schema have '%2' alias" ).arg( actors.size() ).arg( alias ) );
    }
    
    Actor * ret = actors.first();
    attrName = ret->getParamAliases().key( alias );
    return ret;
}

static void setSchemaCMDLineOptions( Schema * schema, int optionsStartAtIdx ) {
    assert( schema != NULL && optionsStartAtIdx > 0 );
    
    QList<StringPair> parameters = AppContext::getCMDLineRegistry()->getParameters();
    int sz = parameters.size();
    for( int i = optionsStartAtIdx; i < sz; ++i ) {
        const StringPair & param = parameters.at(i);
        if( param.first.isEmpty() ) { // TODO: unnamed parameters not supported yet
            continue;
        }
        
        QString paramAlias = param.first;
        QString paramName;
        Actor * actor = findActorByParamAlias( schema, paramAlias, paramName );
        if( actor == NULL ) {
            assert( paramName.isEmpty() );
            log.trace( WorkflowRunFromCMDLineTask::tr( "alias '%1' not set in schema" ).arg( paramAlias ) );
            continue;
        }
        
        Attribute * attr = actor->getParameter( paramName );
        if( attr == NULL ) {
            log.error( WorkflowRunFromCMDLineTask::tr( "actor parameter '%1' not found" ).arg( paramName ) );
            continue;
        }
        
        DataTypeValueFactory * valueFactory = WorkflowEnv::getDataTypeValueFactoryRegistry()->getById( attr->type->getId() );
        if( valueFactory == NULL ) {
            log.error( WorkflowRunFromCMDLineTask::tr( "cannot parse value from '%1'" ).arg( param.second ) );
            continue;
        }
        
        ActorId id = actor->getId();
        QVariant value = valueFactory->getValueFromString( param.second );
        QList<Iteration>::iterator it = schema->iterations.begin();
        while( it != schema->iterations.end() ) { // TODO: make different values for different iterations
            it->cfg[id].insert( paramName, value );
            ++it;
        }
    }
}

QList<Task*> WorkflowRunFromCMDLineTask::onSubTaskFinished( Task* subTask ) {
    assert( subTask != NULL );
    QList<Task*> res;
    
    propagateSubtaskError();
    if( hasErrors() || isCanceled() ) {
        return res;
    }
    assert( !hasErrors() ); // if error, we won't be here
    
    if( loadTask == subTask ) {
        Schema * schema = loadTask->getSchema();
        assert( schema != NULL );
        
        setSchemaCMDLineOptions( schema, optionsStartAt );
        if( schema->domain.isEmpty() ) {
            schema->domain = WorkflowEnv::getDomainRegistry()->getAllIds().value(0);
        }
        res << new WorkflowRunTask( *schema, schema->iterations );
    } else if( qobject_cast<WorkflowRunTask*>( subTask ) == NULL ) {
        assert( false );
    }
    
    return res;
}

QString WorkflowRunFromCMDLineTask::getPathToSchemaFile(const QString & name ) const {
    // full path given
    if( QFile::exists( name ) ) {
        return name;
    }
    // search file as a standard cmdline command
    QString filenameWithCMDPrefix = QString( WORKFLOW_COMMANDS_PATH_PREFIX ) + ":" + name;
    if( QFile::exists( filenameWithCMDPrefix ) ) {
        return filenameWithCMDPrefix;
    }
    QString filenameWithCMDPrefixAndExt = QString( WORKFLOW_COMMANDS_PATH_PREFIX ) + ":" + name + ".uws";
    if( QFile::exists( filenameWithCMDPrefixAndExt ) ) {
        return filenameWithCMDPrefixAndExt;
    }
    // search file in data dirs
    QString filenameWithDataPrefix = QString( PATH_PREFIX_DATA ) + ":" + "cmdline/"+name;
    if( QFile::exists( filenameWithDataPrefix ) ) {
        return filenameWithDataPrefix;
    }
    QString filenameWithDataPrefixAndExt = QString( PATH_PREFIX_DATA ) + ":" + "cmdline/" + name + ".uws";
    if( QFile::exists( filenameWithDataPrefixAndExt ) ) {
        return filenameWithDataPrefixAndExt;
    }
    
    // if no such file found -> search name in settings. user saved schemas
    Settings * settings = AppContext::getSettings();
    assert( settings != NULL );
    QVariantMap pathsMap = settings->getValue( SaveWorkflowTask::SCHEMA_PATHS_SETTINGS_TAG ).toMap();
    QString path = pathsMap.value( name ).toString();
    if( QFile::exists( path ) ) {
        return path;
    }
    
    return QString();
}

}//namespace
