/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "WorkflowUtils.h"

#include <workflow/Descriptor.h>
#include <workflow_support/SchemaSerializer.h>
#include <core_api/GObject.h>
#include <core_api/DocumentModel.h>
#include <document_format/DocumentFormatUtils.h>
#include <util_gui/DialogUtils.h>

#include <QtCore/QDir>
#include <QtCore/QFileInfo>
#include <QtCore/QUrl>
#include <QtGui/QListWidgetItem>

namespace GB2 {

QString DesignerUtils::getRichDoc(const Descriptor& d) {
    return QString("<b>%1</b> : %2").arg(d.getDisplayName()).arg(d.getDocumentation());
}

const QString DesignerUtils::SCHEMA_FILE_EXT("uws");
QString DesignerUtils::getSchemaFileFilter(bool any) {
    return DialogUtils::prepareFileFilter(tr("UGENE workflow schemes"), QStringList(SCHEMA_FILE_EXT), any);
}

QString DesignerUtils::getDropUrl(QList<DocumentFormat*>& fs, const QMimeData* md) {
    QString url;
    const GObjectMimeData* gomd = qobject_cast<const GObjectMimeData*>(md);
    const DocumentMimeData* domd = qobject_cast<const DocumentMimeData*>(md);
    if (gomd) {
        GObject* obj = gomd->objPtr.data();
        if (obj) {
            fs << obj->getDocument()->getDocumentFormat();
            url = obj->getDocument()->getURLString();
        }
    } else if (domd) {
        Document* doc = domd->objPtr.data();
        if (doc) {
            fs << doc->getDocumentFormat();
            url = doc->getURLString();
        }
    } else if (md->hasUrls()) {
        QList<QUrl> urls = md->urls();
        if (urls.size() == 1)
        {
            url = urls.at(0).toLocalFile();
            fs += DocumentFormatUtils::detectFormat(url);
        }
    }
    return url;
}

void DesignerUtils::setQObjectProperties(QObject &o , const QVariantMap & params) {
    QMapIterator<QString, QVariant> i(params);
    while (i.hasNext()) {
        i.next();
        //log.debug("set param " + i.key() + "="+i.value().toString());
        o.setProperty(i.key().toAscii(), i.value());
    }
}

QStringList DesignerUtils::expandToUrls(const QString& s) {
    QStringList urls = s.split(";");
    QStringList result;
    QRegExp wcard("[*?\\[\\]]");
    foreach(QString url, urls) 
    {
        int idx = url.indexOf(wcard);
        if (idx >= 0) {
            int dirIdx = url.lastIndexOf('/', idx);
            QDir dir;
            if (dirIdx >= 0) {
                dir = QDir(url.left(dirIdx));
                url = url.right(url.length() - dirIdx - 1);
            }

            foreach(QFileInfo fi, dir.entryInfoList((QStringList() << url), QDir::Files|QDir::NoSymLinks)) {
                result << fi.absoluteFilePath();
            }
        } else {
            //if (QFile::exists(url)) 
            {
                result << url;
            }
        }
    }
    return result;
}

bool DesignerUtils::validate(const Schema& schema, QList<QListWidgetItem*>* infoList) {
    bool good = true;
    foreach (Actor* a, schema.procs) {
        foreach(Port* p, a->getPorts()) {
            QStringList l;
            bool ag = p->validate(l);
            good &= ag;
            if (infoList && !l.isEmpty()) {
                foreach(QString s, l) {
                    QListWidgetItem* item = new QListWidgetItem(a->getProto()->getIcon(), 
                        QString("%1 : %2").arg(a->getLabel()).arg(s));
                    item->setData(PORT_REF, p->getId());
                    item->setData(ACTOR_REF, a->getId());
                    infoList->append(item);
                }
            }
        }
    }

    foreach (const Iteration& it, schema.iterations) {
        Schema sh;
        QMap<ActorId, ActorId> map = SchemaSerializer::deepCopy(schema, &sh);
        sh.applyConfiguration(it, map);

        foreach (Actor* a, sh.procs) {
            QStringList l;
            bool ag = a->validate(l);
            good &= ag;
            if (infoList && !l.isEmpty()) {
                foreach(QString s, l) {
                    QListWidgetItem* item = new QListWidgetItem(a->getProto()->getIcon(), 
                        tr("Iteration '%3', %1 : %2").arg(a->getLabel()).arg(s).arg(it.name));
                    item->setData(ACTOR_REF, map.key(a->getId()));
                    item->setData(ITERATION_REF, it.id);
                    infoList->append(item);
                }
            }
        }
    }
    return good;
}

bool DesignerUtils::validate(const Schema& schema, QList<QMap<int, QVariant> >* infoList) {
    bool good = true;
    foreach (Actor* a, schema.procs) {
        foreach(Port* p, a->getPorts()) {
            QStringList l;
            bool ag = p->validate(l);
            good &= ag;
            if (infoList && !l.isEmpty()) {
                foreach(QString s, l) {
                    QMap<int, QVariant> item;
                    item[TEXT_REF] = QString("%1 : %2").arg(a->getLabel()).arg(s);
                    item[PORT_REF] = p->getId();
                    item[ACTOR_REF] = a->getId();
                    infoList->append(item);
                }
            }
        }
    }

    foreach (const Iteration& it, schema.iterations) {
        Schema sh;
        QMap<ActorId, ActorId> map = SchemaSerializer::deepCopy(schema, &sh);
        sh.applyConfiguration(it, map);

        foreach (Actor* a, sh.procs) {
            QStringList l;
            bool ag = a->validate(l);
            good &= ag;
            if (infoList && !l.isEmpty()) {
                foreach(QString s, l) {
                    QMap<int, QVariant> item;
                    item[TEXT_REF] = tr("Iteration '%3', %1 : %2").arg(a->getLabel()).arg(s).arg(it.name);
                    item[ACTOR_REF] = map.key(a->getId());
                    item[ITERATION_REF] = it.id;
                    infoList->append(item);
                }
            }
        }
    }
    return good;
}

QVariant PrompterBaseImpl::getParameter(const QString& id) {
    if (map.contains(id)) {
        return map.value(id);
    } else {
        return target->getParameter(id)->value;
    }
}

QString PrompterBaseImpl::getURL(const QString& id, bool * empty ) {
    QString url = getParameter(id).toString();
    if( empty != NULL ) { *empty = false; }
    if (url.isEmpty()) {
        url = "<font color='red'>"+tr("unset")+"</font>";
        if( empty != NULL ) { *empty = true; }
    } else if (url.indexOf(";") != -1) {
        url = tr("the list of files");
    } else {
        url = QFileInfo(url).fileName();
    }
    return url;
}

QString PrompterBaseImpl::getRequiredParam(const QString& id) {
    QString url = getParameter(id).toString();
    if (url.isEmpty()) {
        url = "<font color='red'>"+tr("unset")+"</font>";
    }
    return url;
}

QString PrompterBaseImpl::getScreenedURL(BusPort* input, const QString& id, const QString& slot) {
    bool empty = false;
    QString attrUrl = QString("<u>%1</u>").arg(getURL(id, &empty));
    if( !empty ) {
        return attrUrl;
    }
    
    Actor * origin = input->getProducer( slot );
    QString slotUrl;
    if( origin != NULL ) {
        slotUrl = tr("file(s) alongside of input sources of <u>%1</u>").arg(origin->getLabel());
        return slotUrl;
    }
    
    assert( !attrUrl.isEmpty() );
    return attrUrl;
}

QString PrompterBaseImpl::getProducers( const QString& port, const QString& slot )
{
    BusPort* input = qobject_cast<BusPort*>(target->getPort(port));
    QList<Actor*> producers = input->getProducers(slot);

    QStringList labels;
    foreach(Actor* a, producers) {
        labels << a->getLabel();
    }
    return labels.join(", ");
}
}//namespace
