/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "PWMSearchDialogController.h"
#include "PWMBuildDialogController.h"
#include "PWMJASPARDialogController.h"
#include "WeightMatrixSearchTask.h"
#include "WeightMatrixAlgorithm.h"
#include "WeightMatrixIO.h"

#include <util_ov_annotated_dna/AnnotatedDNAView.h>
#include <util_ov_annotated_dna/ADVSequenceObjectContext.h>

#include <util_weight_matrix/PWMConversionAlgorithmRegistry.h>

#include <util_gui/CreateAnnotationDialog.h>
#include <util_gui/CreateAnnotationWidgetController.h>
#include <util_gui/DialogUtils.h>

#include <gobjects/DNASequenceObject.h>
#include <gobjects/GObjectUtils.h>

#include <core_api/DNATranslation.h>
#include <core_api/DNAAlphabet.h>
#include <core_api/AppContext.h>
#include <core_api/IOAdapter.h>
#include <core_api/L10n.h>
#include <util_gui/GUIUtils.h>
#include <util_text/TextUtils.h>
#include <util_tasks/CreateAnnotationTask.h>
#include <selection/DNASequenceSelection.h>

#include <assert.h>

#include <QtCore/QFileInfo>
#include <QtGui/QMessageBox>
#include <QtGui/QListWidgetItem>
#include <QtGui/QFileDialog>

namespace GB2 {

class WeightMatrixResultItem : public QTreeWidgetItem {
public:
    WeightMatrixResultItem(const WeightMatrixSearchResult& r);
    WeightMatrixSearchResult res;
    virtual bool operator< ( const QTreeWidgetItem & other ) const {
        const WeightMatrixResultItem* o = (const WeightMatrixResultItem*)&other;
        int n = treeWidget()->sortColumn();
        switch (n) {
            case 0 : 
                return res.region.startPos < o->res.region.startPos;
            case 1: 
                return res.complement != o->res.complement ? res.complement :  (res.region.startPos < o->res.region.startPos);
            case 2:
                return res.score < o->res.score;
        }
        return false;
    }
};


/* TRANSLATOR GB2::SiteconSearchDialogController */ 

PWMSearchDialogController::PWMSearchDialogController(ADVSequenceObjectContext* _ctx, QWidget *p):QDialog(p) {
    setupUi(this);
    model = NULL;
    
    //props = _props;
    
    ctx = _ctx;
    task = NULL;
    
    initialSelection = ctx->getSequenceSelection()->isEmpty() ? LRegion() : ctx->getSequenceSelection()->getSelectedRegions().first();
    int seqLen = ctx->getSequenceLen();
    sbRangeStart->setMinimum(1);
    sbRangeStart->setMaximum(seqLen);
    sbRangeEnd->setMinimum(1);
    sbRangeEnd->setMaximum(seqLen);
    
    sbRangeStart->setValue(1);
    sbRangeEnd->setValue(seqLen);

    connectGUI();
    updateState();
    
    scoreValueLabel->setText(QString("%1%").arg(scoreSlider->value()));

    QStringList algo = AppContext::getPWMConversionAlgorithmRegistry()->getAlgorithmIds();
    algorithmCombo->addItems(algo);

    pbSelectModelFile->setFocus();
    timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), SLOT(sl_onTimer()));
}

PWMSearchDialogController::~PWMSearchDialogController() {
    if (model!=NULL) {
        delete model;
        model = NULL;
    }
}

void PWMSearchDialogController::connectGUI() {
    //buttons
    connect(pbSelectModelFile, SIGNAL(clicked()), SLOT(sl_selectModelFile()));
    connect(pbSaveAnnotations, SIGNAL(clicked()), SLOT(sl_onSaveAnnotations()));
    connect(pbClear, SIGNAL(clicked()), SLOT(sl_onClearList()));
    connect(pbSearch, SIGNAL(clicked()), SLOT(sl_onSearch()));
    connect(pbClose, SIGNAL(clicked()), SLOT(sl_onClose()));

    connect(scoreSlider, SIGNAL(valueChanged(int)), SLOT(sl_onSliderMoved(int)));
    connect(buildButton, SIGNAL(clicked()), SLOT(sl_onBuildMatrix()));
    connect(jasparButton, SIGNAL(clicked()), SLOT(sl_onSearchJaspar()));
    connect(algorithmCombo, SIGNAL(currentIndexChanged(QString)), SLOT (sl_onAlgoChanged(QString)));

    //results list
    connect(resultsTree, SIGNAL(itemActivated(QTreeWidgetItem*, int)), SLOT(sl_onResultActivated(QTreeWidgetItem*, int)));

    resultsTree->installEventFilter(this);
}


void PWMSearchDialogController::updateState() {
    bool hasInitialSelection = initialSelection.len > 0;
    rbSelectionRange->setEnabled(hasInitialSelection);

    bool hasActiveTask = task!=NULL;
    bool hasCompl = ctx->getComplementTT()!=NULL;

    bool hasResults = resultsTree->topLevelItemCount() > 0;
    
    pbSearch->setEnabled(!hasActiveTask);
    
    pbSaveAnnotations->setEnabled(!hasActiveTask && hasResults);
    pbClear->setEnabled(!hasActiveTask && hasResults);
    pbClose->setText(hasActiveTask ? tr("Cancel") : tr("Close"));  

    rbBoth->setEnabled(!hasActiveTask && hasCompl);
    rbDirect->setEnabled(!hasActiveTask);
    rbComplement->setEnabled(!hasActiveTask && hasCompl);
    
    updateStatus();
}

void PWMSearchDialogController::updateStatus() {
    QString message;
    if (task != NULL) {
        message = tr("Progress %1% ").arg(qMax(0, task->getProgress()));
    }
    message += tr("Results found: %1.").arg(resultsTree->topLevelItemCount());
    statusLabel->setText(message);
}

bool PWMSearchDialogController::eventFilter(QObject *obj, QEvent *ev) {
    if (obj == resultsTree && ev->type() == QEvent::KeyPress) {
        QKeyEvent* ke = (QKeyEvent*)ev;
        if (ke->key() == Qt::Key_Space) {
            WeightMatrixResultItem* item = static_cast<WeightMatrixResultItem*>(resultsTree->currentItem());
            if (item != NULL) {
                sl_onResultActivated(item, 0);
            }
        }
    }
    return false;
}

void PWMSearchDialogController::sl_selectModelFile() {
    LastOpenDirHelper lod(WeightMatrixIO::WEIGHT_MATRIX_ID);
    lod.url = QFileDialog::getOpenFileName(this, tr("Select file with frequency or weight matrix"), lod, WeightMatrixIO::getPFMFileFilter(false) + "\n" + WeightMatrixIO::getPWMFileFilter());
    if (lod.url.isEmpty()) {
        return;
    }
    
    TaskStateInfo siPFM;
    IOAdapterFactory* iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::url2io(lod.url));
    PWMatrix m;
    PFMatrix temp = WeightMatrixIO::readPFMatrix(iof, lod.url, siPFM);
    if (siPFM.hasErrors()) {
        TaskStateInfo siPWM;    
        m = WeightMatrixIO::readPWMatrix(iof, lod.url, siPWM);
        if (siPWM.hasErrors()) {
            QMessageBox::critical(this, L10N::errorTitle(), siPWM.getError());
            return;
        }
        algorithmLabel->setEnabled(false);
        algorithmCombo->setEnabled(false);
    } else {
        algorithmLabel->setEnabled(true);
        algorithmCombo->setEnabled(true);
        PWMConversionAlgorithmFactory* factory = AppContext::getPWMConversionAlgorithmRegistry()->getAlgorithmFactory(algorithmCombo->currentText());
        PWMConversionAlgorithm* algo = factory->createAlgorithm();
        m = algo->convert(temp);
    }
    updateModel(m);
    QFileInfo fi(lod.url);
    modelFileEdit->setText(fi.absoluteFilePath());
}

void PWMSearchDialogController::updateModel(const PWMatrix& m) {
    if (model!=NULL) {
        delete model;
        model = NULL;
    }
    model = new PWMatrix();
    *model = m;
    scoreSlider->setValue(75);
}

void PWMSearchDialogController::sl_onSaveAnnotations() {
    if (resultsTree->topLevelItemCount() == 0) {
        return;
    }
    
    CreateAnnotationModel m;
    m.sequenceObjectRef = ctx->getSequenceObject();
    m.hideLocation = true;
    m.sequenceLen = ctx->getSequenceObject()->getSequenceLen();
    CreateAnnotationDialog d(this, m);
    int rc = d.exec();
    if (rc != QDialog::Accepted) {
        return;
    }
    const QString& name = m.data->name;
    QList<SharedAnnotationData> list;
    for (int i=0, n = resultsTree->topLevelItemCount(); i<n; ++i) {
        WeightMatrixResultItem* item = static_cast<WeightMatrixResultItem* >(resultsTree->topLevelItem(i));
        list.append(item->res.toAnnotation(name));
    }

    CreateAnnotationsTask* t = new CreateAnnotationsTask(m.getAnnotationObject(), m.groupName, list);
    AppContext::getTaskScheduler()->registerTopLevelTask(t);
}

void PWMSearchDialogController::sl_onClearList() {
    resultsTree->clear();
    updateState();
}

void PWMSearchDialogController::sl_onSearch() {
    runTask();
}

void PWMSearchDialogController::reject() {
    if (task!=NULL) {
        task->cancel();
        return;
    }
    QDialog::reject();
}


void PWMSearchDialogController::sl_onClose() {
    reject();
}

void PWMSearchDialogController::sl_onSliderMoved(int value) {
    scoreValueLabel->setText(QString("%1%").arg(value));
    updateState();
}

void PWMSearchDialogController::sl_onBuildMatrix() {
    PWMBuildDialogController bd(this);
    if (bd.exec() == QDialog::Accepted) {
        TaskStateInfo siPFM;
        IOAdapterFactory* iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::url2io(bd.outputEdit->text()));
        PWMatrix m;
        PFMatrix temp = WeightMatrixIO::readPFMatrix(iof, bd.outputEdit->text(), siPFM);
        if (siPFM.hasErrors()) {
            TaskStateInfo siPWM;    
            m = WeightMatrixIO::readPWMatrix(iof, bd.outputEdit->text(), siPWM);
            if (siPWM.hasErrors()) {
                QMessageBox::critical(this, L10N::errorTitle(), siPWM.getError());
                return;
            }
            algorithmLabel->setEnabled(false);
            algorithmCombo->setEnabled(false);
        } else {
            algorithmLabel->setEnabled(true);
            algorithmCombo->setEnabled(true);
            PWMConversionAlgorithmFactory* factory = AppContext::getPWMConversionAlgorithmRegistry()->getAlgorithmFactory(algorithmCombo->currentText());
            PWMConversionAlgorithm* algo = factory->createAlgorithm();
            m = algo->convert(temp);
        }
        updateModel(m);
        QFileInfo fi(bd.outputEdit->text());
        modelFileEdit->setText(fi.absoluteFilePath());
    }
}

void PWMSearchDialogController::sl_onAlgoChanged(QString newAlgo){
    if (modelFileEdit->text().isEmpty()) return;
    TaskStateInfo siPFM;
    IOAdapterFactory* iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::url2io(modelFileEdit->text()));
    PWMatrix m;
    PFMatrix temp = WeightMatrixIO::readPFMatrix(iof, modelFileEdit->text(), siPFM);
    if (siPFM.hasErrors()) {
        TaskStateInfo siPWM;    
        m = WeightMatrixIO::readPWMatrix(iof, modelFileEdit->text(), siPWM);
        if (siPWM.hasErrors()) {
            QMessageBox::critical(this, L10N::errorTitle(), siPWM.getError());
            return;
        }
        algorithmLabel->setEnabled(false);
        algorithmCombo->setEnabled(false);
    } else {
        algorithmLabel->setEnabled(true);
        algorithmCombo->setEnabled(true);
        PWMConversionAlgorithmFactory* factory = AppContext::getPWMConversionAlgorithmRegistry()->getAlgorithmFactory(algorithmCombo->currentText());
        PWMConversionAlgorithm* algo = factory->createAlgorithm();
        m = algo->convert(temp);
    }
    updateModel(m);
    QFileInfo fi(modelFileEdit->text());
    modelFileEdit->setText(fi.absoluteFilePath());
}

void PWMSearchDialogController::sl_onSearchJaspar() {
    PWMJASPARDialogController jd(this);
    if (jd.exec() == QDialog::Accepted) {
        IOAdapterFactory* iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::url2io(jd.fileName));
        TaskStateInfo siPFM;
        PWMatrix m;
        PFMatrix temp = WeightMatrixIO::readPFMatrix(iof, jd.fileName, siPFM);
        if (siPFM.hasErrors()) {
            TaskStateInfo siPWM;    
            m = WeightMatrixIO::readPWMatrix(iof, jd.fileName, siPWM);
            if (siPWM.hasErrors()) {
                QMessageBox::critical(this, L10N::errorTitle(), siPWM.getError());
                return;
            }
            algorithmLabel->setEnabled(false);
            algorithmCombo->setEnabled(false);
        } else {
            algorithmLabel->setEnabled(true);
            algorithmCombo->setEnabled(true);
            PWMConversionAlgorithmFactory* factory = AppContext::getPWMConversionAlgorithmRegistry()->getAlgorithmFactory(algorithmCombo->currentText());
            PWMConversionAlgorithm* algo = factory->createAlgorithm();
            m = algo->convert(temp);
        }
        updateModel(m);
        QFileInfo fi(jd.fileName);
        modelFileEdit->setText(fi.canonicalFilePath());
    }
}

void PWMSearchDialogController::runTask() {
    assert(task == NULL);
    if (model == NULL) {
        QMessageBox::critical(this, L10N::errorTitle(), tr("Model not selected"));
        modelFileEdit->setFocus();
        return;
    }
    if (model->getLength() == 0) {
        QMessageBox::critical(this, L10N::errorTitle(), tr("Zero length or corrupted model"));
        modelFileEdit->setFocus();
        return;
    }
    LRegion reg;
    if (rbSequenceRange->isChecked()) {
        reg = ctx->getSequenceObject()->getSequenceRange();
    } else if (rbSelectionRange->isChecked()) {
        reg = initialSelection;
    } else {
        reg.startPos = sbRangeStart->value();
        reg.len = sbRangeEnd->value() - sbRangeStart->value() + 1;
        if (reg.len <= model->getLength()) {
            QMessageBox::critical(this, L10N::errorTitle(), tr("Range is too small"));
            sbRangeEnd->setFocus();
            return;
        }
    }
    const char* seq = ctx->getSequenceData().constData() + reg.startPos;
    int len = reg.len;

    WeightMatrixSearchCfg cfg;
    cfg.complTT = rbBoth->isChecked() || rbComplement->isChecked() ? ctx->getComplementTT() : NULL;
    cfg.complOnly = rbComplement->isChecked();
    cfg.minPSUM = scoreSlider->value();
    
    //TODO: ask if to clear
    sl_onClearList();

    task = new WeightMatrixSearchTask(*model, seq, len, cfg, reg.startPos);
    connect(task, SIGNAL(si_stateChanged()), SLOT(sl_onTaskFinished()));
    AppContext::getTaskScheduler()->registerTopLevelTask(task);
    updateState();
    timer->start(400);
}

void PWMSearchDialogController::sl_onTaskFinished() {
    task = qobject_cast<WeightMatrixSearchTask*>(sender());
    if (!task->isFinished()) {
        return;
    }
    timer->stop();
    importResults();
    task = NULL;
    updateState();
}

void PWMSearchDialogController::sl_onTimer() {
    importResults();
}

void PWMSearchDialogController::importResults() {
    resultsTree->setSortingEnabled(false);
    
    QList<WeightMatrixSearchResult> newResults = task->takeResults();
    foreach(const WeightMatrixSearchResult& r, newResults) {
        WeightMatrixResultItem* item  = new WeightMatrixResultItem(r);
        resultsTree->addTopLevelItem(item);
    }
    updateStatus();
    
    resultsTree->setSortingEnabled(true);
}

void PWMSearchDialogController::sl_onResultActivated(QTreeWidgetItem* i, int col) {
    Q_UNUSED(col);
    assert(i);
    WeightMatrixResultItem* item = static_cast<WeightMatrixResultItem*>(i);
    DNASequenceSelection* sel = ctx->getSequenceSelection();
    sel->clear();
    sel->addRegion(item->res.region);
}


//////////////////////////////////////////////////////////////////////////
/// tree

WeightMatrixResultItem::WeightMatrixResultItem(const WeightMatrixSearchResult& r) : res(r) 
{
    QString range = QString("%1..%2").arg(r.region.startPos + 1).arg(r.region.endPos());
    setTextAlignment(0, Qt::AlignRight);
    setTextAlignment(1, Qt::AlignRight);
    setTextAlignment(2, Qt::AlignRight);

    setText(0, range);
    QString strand = res.complement ? PWMSearchDialogController::tr("Complement strand") : PWMSearchDialogController::tr("Direct strand") ;
    setText(1, strand);
    setText(2, QString::number(res.score, 'f', 2)+"%");
}


}//namespace
