/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "Muscle4Worker.h"
#include "Muscle4TaskLocalStorage.h"

#include <workflow/IntegralBusModel.h>
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowRegistry.h>
#include <workflow_support/CoreDataTypes.h>
#include <workflow_library/BioDatatypes.h>
#include <workflow_library/BioActorLibrary.h>
#include <workflow_support/DelegateEditors.h>
#include <workflow_support/CoreLibConstants.h>
#include <core_api/Log.h>

namespace GB2 {
namespace LocalWorkflow {

const QString Muscle4WorkerFactory::ACTOR_ID("muscle4.align");

static LogCategory log(ULOG_CAT_WD);

void Muscle4WorkerFactory::init() {
    QList<PortDescriptor*> p; QList<Attribute*> a;
    Descriptor ind(CoreLibConstants::IN_PORT_ID, Muscle4Worker::tr("Input MSA"), Muscle4Worker::tr("Multiple sequence alignment to be processed."));
    Descriptor oud(CoreLibConstants::OUT_PORT_ID, Muscle4Worker::tr("Multiple sequence alignment"), Muscle4Worker::tr("Result of alignment."));
    p << new PortDescriptor(ind, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), true /*input*/);
    p << new PortDescriptor(oud, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), false /*input*/, true /*multi*/);
    
	Descriptor desc(ACTOR_ID, GB2::LocalWorkflow::Muscle4Worker::tr("MUSCLE 4 multiple alignment"), 
        Muscle4Worker::tr("MUSCLE is public domain multiple alignment software for protein and nucleotide sequences."
        "<p><dfn>MUSCLE stands for MUltiple Sequence Comparison by Log-Expectation.</dfn></p>"));
    ActorPrototype* proto = new BusActorPrototype(desc, p, a);

    QMap<QString, PropertyDelegate*> delegates;    

    proto->setEditor(new DelegateEditor(delegates));
    proto->setPrompter(new Muscle4Prompter());
    proto->setIconPath(":umuscle4/images/umuscle4_16.png");
    WorkflowEnv::getProtoRegistry()->registerProto(BioActorLibrary::CATEGORY_ALIGNMENT(), proto);

    DomainFactory* localDomain = WorkflowEnv::getDomainRegistry()->getById(LocalDomainFactory::ID);
    localDomain->registerEntry(new Muscle4WorkerFactory());
}

QString Muscle4Prompter::composeRichDoc() {
    BusPort* input = qobject_cast<BusPort*>(target->getPort(CoreLibConstants::IN_PORT_ID));
    Actor* producer = input->getProducer(CoreLibConstants::IN_PORT_ID);
    QString producerName = producer ? tr(" from %1").arg(producer->getLabel()) : "";
    
    QString doc = tr("For each MSA<u>%1</u>, build the alignment using <u>\"Default\" preset</u> and send it to output.");
    return doc;
}

void Muscle4Worker::init() {
    input = ports.value(CoreLibConstants::IN_PORT_ID);
    output = ports.value(CoreLibConstants::OUT_PORT_ID);
}

bool Muscle4Worker::isReady() {
    return (input && input->hasMessage());
}

Task* Muscle4Worker::tick() {
    MAlignment msa = input->get().getData().value<MAlignment>();
	Task* t = new Muscle4Task(msa);
    connect(t, SIGNAL(si_stateChanged()), SLOT(sl_taskFinished()));
    return t;
}

void Muscle4Worker::sl_taskFinished() {
    Muscle4Task* t = qobject_cast<Muscle4Task*>(sender());
	if (t->getState() != Task::State_Finished) {
		return;
	}
    
    QVariant v = qVariantFromValue<MAlignment>(t->resultMA);
    output->put(Message(BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), v));
    
	if (input->isEnded()) {
        output->setEnded();
    }
    log.info(tr("Aligned %1 with MUSCLE4").arg(t->resultMA.getName()));
}

bool Muscle4Worker::isDone() {
    return !input || input->isEnded();
}

} //namespace LocalWorkflow
} //namespace GB2
